-- Table driven CRC routines                                      G.Burke 12/01


-------------------------------------------------------------------------------
--  This library includes routines for generating CRC lookup tables, and     --
--    calculating the CRC checksum from a given file or data chunk.          --
-------------------------------------------------------------------------------
-- I highly reccommend "A PAINLESS GUIDE TO CRC ERROR DETECTION ALGORITHMS"  --
--  by Ross N. Williams. This is a great read for anyone interesterd in the  --
--              subject and is redily available on-line.                     --
-------------------------------------------------------------------------------

--                 ***Freeware*** Use as you see fit.

-------------------------------------------------------------------------------
-- global function crc(sequence data,sequence spec)
-------------------------------------------------------------------------------

-- returns the crc checksum for data.
-- data must be a 1D sequence of 8-bit integers (bytes)
-- should return 0 if data already has checksum appended to it

--i.e.

--  printf(1,"%x\n",    crc("123456789",CRC16)      )

--          or

--  D= crc(MyData,CRC32)        


-------------------------------------------------------------------------------
-- global function file_crc(sequence path,sequence spec)
-------------------------------------------------------------------------------

--returns the crc checksum for the specified file

-- i.e.
-- check = file_crc("c:\\myFile.dat",CRC32)

-------------------------------------------------------------------------------
--global function CRCgetTable(object poly,integer w)
-------------------------------------------------------------------------------
-- Returns generates and returns a lookup table for the specified poly 
-- w is the width in bits. Each entry in the table returned will be a 
-- sequence of bytes of a length appropriate to the width of the poly.

--i.e.

--  s=CRCgetTable(#04C11DB7,32)

-- *This is done automatically by the crc() and file_crc() routines above

-------------------------------------------------------------------------------





-------------------
-------------------
---- CRC Specs ----
-------------------
-------------------

-- CRC Specifications have been categorized as follows:
--  (see the pre-defined constants below)

--  {   Name
--      Width in bits (not including implicit '1' bit)
--      Poly 
--      Initial register value
--      Reflect Input (bool)
--      Reflect Output (bool)
--      XOR Out Value           
--      check value for the string "123456789"  }

global constant 
	CRC_True=1,
	CRC_False=0,
	
	
	
	CRCspecs={

	{   "CRC-16",-- also ARC
	      16,
	    #8005,
	    #0000,
	     CRC_True,
	     CRC_True,
	    #0000,
	    #BB3D
	},
	
	{   "CRC-32", -- also pkzip
	      32,
	    #04C11DB7,
	    #FFFFFFFF,
	     CRC_True,
	     CRC_True,
	    #FFFFFFFF,
	    #CBF43926
	},
	
	{   "CRC-16/CITT",
	      16,
	    #1024,
	    #FFFF,
	     CRC_False,
	     CRC_False,
	    #0000,
	    #31B8
	}},
	
	
	CRC16=CRCspecs[1],
	CRC32=CRCspecs[2]



global constant CRC_Name=1,
		CRC_Width=2,
		CRC_Poly=3,
		CRC_Init=4,
		CRC_RefIn=5,
		CRC_RefOut=6,
		CRC_XOROut=7


include file.e

global function CRCflen(integer fn)
    --return length in bytes of open file fn
    atom p,r
    p=where(fn)
    r=seek(fn,-1)
    r=where(fn)
    p=seek(fn,p)
    return r
end function

function CRCgetBytes(atom a)
    --split an atom into a sequence of bytes
    sequence r
    r={}
    while a do
	r=and_bits(255,a)&r
	a=floor(a/256)
    end while
    return r
end function

function CRCgetAtom(sequence s)
    --make an atom from sequence of bytes
    -- (reverse of above)
    atom r
    r=0
    for x=1 to length(s) do
	r+=s[x]*power(256,length(s)-x)
    end for
    return r
end function

function top_bit(atom a)
    -- return index to leftmost '1' bit in a
    integer r
    r=0
    while a do
	r+=1
	a=floor(a/2)
    end while
    return r
end function

function LShift(sequence s,integer n)
    --shift a sequence of bytes left by one bit
    --repeat n times
    for b=1 to n do
	s=and_bits(s*2,255)+((and_bits(s[2..length(s)],128)!=0)&0)
    end for
    return s
end function

-- function RShift(sequence s)
--  --  *Not Required*
--  --shift a sequence of bytes right by one bit
--  sequence carry
--  for b=1 to n do
--      s=floor(s/2)+(0&and_bits(s[1..length(s)-1],1))*128
--  end for
--  return s
-- end function

global function CRCgetTable(object poly,integer w)
    -- generate crc lookup table
    sequence buff,r,pshifted
    integer nBytes,shift
    
    r=repeat({},256)--container for table
    
    if atom(poly) then
	--make poly a string of bytes
	--(4 bytes for crc32, 2 bytes for crc16 etc..)
	poly=CRCgetBytes(poly)
    end if
    
    nBytes=floor(w/8)+(remainder(w,8)>0)
    poly=repeat(0,nBytes-length(poly))&poly
    poly=1&poly--prepend implicit '1' bit
    -- ( have to add another byte -> 00000001 )
    
    -- make mini table of Left-Shifted poly values
    -- ie string of bytes equal to poly lshifted from 0 to 7 times
    pshifted=repeat(poly,8)
    for x=2 to 8 do
	pshifted[x]=LShift(pshifted[x-1],1)
    end for
    
    -- for each possible 8-bit value
    for x=0 to 255 do
	-- make a buffer nBytes+1 long
	-- put x in top byte
	buff=x&repeat(0,nBytes)
	
	--while any '1' bits remain in the top byte of the buffer
	--XOR the whole buffer with the appropriate shifted poly
	--from the mini table above
	shift=top_bit(x)
	while shift do
	    buff=xor_bits(buff,pshifted[shift])
	    shift=top_bit(buff[1])
	end while
	
	--table entry value is in remainder of buffer
	r[x+1]=buff[2..length(buff)]
    end for
    
    return r --return table
end function

 
sequence curtable,      -- poly for current lookup table
	 lutable        -- container for lookup table

curtable={}
lutable={}

global procedure CRCsetTable(object poly,integer w)
    -- set current table (dont remake if same poly)
    if atom(poly) then
	poly=CRCgetBytes(poly)
    end if
    if compare(poly,curtable) then
	curtable=poly
	lutable=CRCgetTable(poly,w)
    end if
end procedure 

constant -- helpers for building reflect table
    refli={128,64,32,16,8,4,2,1},
    reflo={1,2,4,8,16,32,64,128}

function do_ref(integer i)
    --reflect an 8 bit integer
    integer r
    r=0
    for x=1 to 8 do
	r+=reflo[x]*(and_bits(i,refli[x])!=0)
    end for
    return r
end function

function getReflectTable()
    --generate lookup for reflecting bytes
    sequence r
    r=repeat(0,255)
    for x=1 to 255 do
	if not r[x] then
	    r[x]=do_ref(x)
	    r[r[x]]=x
	end if
    end for
    return  0&r
end function

constant reftable=getReflectTable()

function reflectBytes(sequence s)
    -- reflect a sequence of bytes
    sequence r
    r={}
    for x=1 to length(s) do
	r=reftable[s[x]+1]&r
    end for
    return r
end function

global function crc(sequence data,sequence spec)
    -- return the checksum for data 
    integer w,wb
    sequence buff,p,xorout
    
    w=                  spec[CRC_Width]     --width in bits
    wb=     floor(w/8)+(remainder(w,8)>0)   --width in whole bytes
    buff=   CRCgetBytes(spec[CRC_Init])     --\
    buff=   repeat(0,wb-length(buff))&buff  ---initialize buffer 
    p=      CRCgetBytes(spec[CRC_Poly])     --convert poly to string of bytes
    p=      repeat(0,wb-length(p))&p
    xorout= CRCgetBytes(spec[CRC_XOROut])   --convert xor out value to bytes
    xorout= repeat(0,wb-length(xorout))&xorout
    
    CRCsetTable(p,w)
    
    -- shift buffer 8 bits left
	
    -- xor ejected byte with next data byte
    -- use as reference to lookup table
    -- * Reflect the data byte if appropriate
    
    -- xor entire buffer with value from table
    
    if spec[CRC_RefIn] then
	for x=1 to length(data) do
	    
	    buff=xor_bits(
		    buff[2..length(buff)]&0,
		    lutable[xor_bits(buff[1],reftable[data[x]+1])+1]    )
	end for
    else
	for x=1 to length(data) do
	    
	    buff=xor_bits(
		    buff[2..length(buff)]&0,
		    lutable[xor_bits(buff[1],data[x])+1]    )
	end for
    end if
    
    -- reflect output buffer if applicable
    if spec[CRC_RefOut] then
	buff=reflectBytes(buff)
    end if
    -- apply the xor-out value
    buff=xor_bits(buff,xorout)
    
    return CRCgetAtom(buff)
end function

global function file_crc(sequence path,sequence spec)
    --identical to crc() only this one reads bytes from a file
    integer w,wb,fn,l
    sequence buff,p,xorout

    fn=open(path,"rb")
    if fn<2 then
	return {}
    end if
    l=CRCflen(fn)
    w=                  spec[CRC_Width]
    wb=     floor(w/8)+(remainder(w,8)>0)
    buff=   CRCgetBytes(spec[CRC_Init])
    buff=   repeat(0,wb-length(buff))&buff
    p=      CRCgetBytes(spec[CRC_Poly])
    p=      repeat(0,wb-length(p))&p
    xorout= CRCgetBytes(spec[CRC_XOROut])
    xorout= repeat(0,wb-length(xorout))&xorout
    CRCsetTable(p,w)
    
    if spec[CRC_RefIn] then
	for x=1 to l do
	    buff=xor_bits(
		    buff[2..length(buff)]&0,
		    lutable[xor_bits(buff[1],reftable[getc(fn)+1])+1]   )
	end for
    else
	for x=1 to l do
	    buff=xor_bits(
		    buff[2..length(buff)]&0,
		    lutable[xor_bits(buff[1],getc(fn))+1]   )
	end for
    end if
    
    close(fn)
    
    if spec[CRC_RefOut] then
	buff=reflectBytes(buff)
    end if
    buff=xor_bits(buff,xorout)
    return CRCgetAtom(buff)
end function

